from __future__ import annotations

import datetime
import os
import re
import shutil
import sys
from io import BytesIO
from pathlib import Path
from typing import Literal, cast

import pytest

from PIL import Image, ImageMode, ImageWin, features

from .helper import (
    assert_image,
    assert_image_equal,
    assert_image_similar,
    assert_image_similar_tofile,
    hopper,
    is_pypy,
)

try:
    from PIL import ImageCms
    from PIL.ImageCms import ImageCmsProfile

    ImageCms.core.profile_open
except ImportError:
    # Skipped via setup_module()
    pass

TYPE_CHECKING = False
if TYPE_CHECKING:
    from typing import Any

SRGB = "Tests/icc/sRGB_IEC61966-2-1_black_scaled.icc"
HAVE_PROFILE = os.path.exists(SRGB)


def setup_module() -> None:
    try:
        from PIL import ImageCms

        # need to hit getattr to trigger the delayed import error
        ImageCms.core.profile_open
    except ImportError as v:
        pytest.skip(str(v))


def skip_missing() -> None:
    if not HAVE_PROFILE:
        pytest.skip("SRGB profile not available")


def test_sanity() -> None:
    # basic smoke test.
    # this mostly follows the cms_test outline.

    # internal version number
    version = features.version_module("littlecms2")
    assert version is not None
    assert re.search(r"\d+\.\d+(\.\d+)?$", version)

    skip_missing()
    i = ImageCms.profileToProfile(hopper(), SRGB, SRGB)
    assert i is not None
    assert_image(i, "RGB", (128, 128))

    i = hopper()
    ImageCms.profileToProfile(i, SRGB, SRGB, inPlace=True)
    assert_image(i, "RGB", (128, 128))

    t = ImageCms.buildTransform(SRGB, SRGB, "RGB", "RGB")
    i = ImageCms.applyTransform(hopper(), t)
    assert i is not None
    assert_image(i, "RGB", (128, 128))

    with hopper() as i:
        t = ImageCms.buildTransform(SRGB, SRGB, "RGB", "RGB")
        ImageCms.applyTransform(hopper(), t, inPlace=True)
        assert i is not None
        assert_image(i, "RGB", (128, 128))

    p = ImageCms.createProfile("sRGB")
    o = ImageCms.getOpenProfile(SRGB)
    t = ImageCms.buildTransformFromOpenProfiles(p, o, "RGB", "RGB")
    i = ImageCms.applyTransform(hopper(), t)
    assert i is not None
    assert_image(i, "RGB", (128, 128))

    t = ImageCms.buildProofTransform(SRGB, SRGB, SRGB, "RGB", "RGB")
    assert t.inputMode == "RGB"
    assert t.outputMode == "RGB"
    i = ImageCms.applyTransform(hopper(), t)
    assert i is not None
    assert_image(i, "RGB", (128, 128))

    # test PointTransform convenience API
    hopper().point(t)


def test_flags() -> None:
    assert ImageCms.Flags.NONE.value == 0
    assert ImageCms.Flags.GRIDPOINTS(0) == ImageCms.Flags.NONE
    assert ImageCms.Flags.GRIDPOINTS(256) == ImageCms.Flags.NONE

    assert ImageCms.Flags.GRIDPOINTS(255) == (255 << 16)
    assert ImageCms.Flags.GRIDPOINTS(-1) == ImageCms.Flags.GRIDPOINTS(255)
    assert ImageCms.Flags.GRIDPOINTS(511) == ImageCms.Flags.GRIDPOINTS(255)


def test_name() -> None:
    skip_missing()
    # get profile information for file
    assert (
        ImageCms.getProfileName(SRGB).strip()
        == "IEC 61966-2-1 Default RGB Colour Space - sRGB"
    )


def test_info() -> None:
    skip_missing()
    assert ImageCms.getProfileInfo(SRGB).splitlines() == [
        "sRGB IEC61966-2-1 black scaled",
        "",
        "Copyright International Color Consortium, 2009",
        "",
    ]


def test_copyright() -> None:
    skip_missing()
    assert (
        ImageCms.getProfileCopyright(SRGB).strip()
        == "Copyright International Color Consortium, 2009"
    )


def test_manufacturer() -> None:
    skip_missing()
    assert ImageCms.getProfileManufacturer(SRGB).strip() == ""


def test_model() -> None:
    skip_missing()
    assert (
        ImageCms.getProfileModel(SRGB).strip()
        == "IEC 61966-2-1 Default RGB Colour Space - sRGB"
    )


def test_description() -> None:
    skip_missing()
    assert (
        ImageCms.getProfileDescription(SRGB).strip() == "sRGB IEC61966-2-1 black scaled"
    )


def test_intent() -> None:
    skip_missing()
    assert ImageCms.getDefaultIntent(SRGB) == 0
    support = ImageCms.isIntentSupported(
        SRGB, ImageCms.Intent.ABSOLUTE_COLORIMETRIC, ImageCms.Direction.INPUT
    )
    assert support == 1


def test_profile_object() -> None:
    # same, using profile object
    p = ImageCms.createProfile("sRGB")
    # assert ImageCms.getProfileName(p).strip() == "sRGB built-in - (lcms internal)"
    # assert ImageCms.getProfileInfo(p).splitlines() ==
    #     ["sRGB built-in", "", "WhitePoint : D65 (daylight)", "", ""]
    assert ImageCms.getDefaultIntent(p) == 0
    support = ImageCms.isIntentSupported(
        p, ImageCms.Intent.ABSOLUTE_COLORIMETRIC, ImageCms.Direction.INPUT
    )
    assert support == 1


def test_extensions() -> None:
    # extensions

    with Image.open("Tests/images/rgb.jpg") as i:
        p = ImageCms.getOpenProfile(BytesIO(i.info["icc_profile"]))
    assert (
        ImageCms.getProfileName(p).strip()
        == "IEC 61966-2.1 Default RGB colour space - sRGB"
    )


def test_exceptions() -> None:
    # Test mode mismatch
    psRGB = ImageCms.createProfile("sRGB")
    pLab = ImageCms.createProfile("LAB")
    t = ImageCms.buildTransform(pLab, psRGB, "LAB", "RGB")
    with pytest.raises(ValueError, match="mode mismatch"):
        t.apply(hopper("RGBA"))
    with pytest.raises(ValueError, match="mode mismatch"):
        t.apply(hopper("LAB"), hopper("RGBA"))
    with pytest.raises(ValueError, match="mode mismatch"):
        t.apply_in_place(hopper("RGBA"))

    # the procedural pyCMS API uses PyCMSError for all sorts of errors
    with hopper() as im:
        with pytest.raises(ImageCms.PyCMSError, match="cannot open profile file"):
            ImageCms.profileToProfile(im, "foo", "bar")

    with pytest.raises(ImageCms.PyCMSError, match="cannot open profile file"):
        ImageCms.buildTransform("foo", "bar", "RGB", "RGB")

    with pytest.raises(ImageCms.PyCMSError, match="Invalid type for Profile"):
        ImageCms.getProfileName(None)  # type: ignore[arg-type]
    skip_missing()

    with pytest.raises(
        ImageCms.PyCMSError,
        match="'NoneType' object cannot be interpreted as an integer",
    ):
        ImageCms.isIntentSupported(SRGB, None, None)  # type: ignore[arg-type]


def test_display_profile() -> None:
    # try fetching the profile for the current display device
    ImageCms.get_display_profile()

    if sys.platform == "win32":
        ImageCms.get_display_profile(ImageWin.HDC(0))
        ImageCms.get_display_profile(ImageWin.HWND(0))


def test_lab_color_profile() -> None:
    ImageCms.createProfile("LAB", 5000)
    ImageCms.createProfile("LAB", 6500)


def test_unsupported_color_space() -> None:
    with pytest.raises(
        ImageCms.PyCMSError,
        match=re.escape(
            "Color space not supported for on-the-fly profile creation (unsupported)"
        ),
    ):
        ImageCms.createProfile("unsupported")  # type: ignore[arg-type]


def test_invalid_color_temperature() -> None:
    with pytest.raises(
        ImageCms.PyCMSError,
        match='Color temperature must be numeric, "invalid" not valid',
    ):
        ImageCms.createProfile("LAB", "invalid")  # type: ignore[arg-type]


@pytest.mark.parametrize("flag", ("my string", -1))
def test_invalid_flag(flag: str | int) -> None:
    with hopper() as im:
        with pytest.raises(
            ImageCms.PyCMSError, match="flags must be an integer between 0 and "
        ):
            ImageCms.profileToProfile(im, "foo", "bar", flags=flag)  # type: ignore[arg-type]


def test_simple_lab() -> None:
    i = Image.new("RGB", (10, 10), (128, 128, 128))

    psRGB = ImageCms.createProfile("sRGB")
    pLab = ImageCms.createProfile("LAB")
    t = ImageCms.buildTransform(psRGB, pLab, "RGB", "LAB")

    i_lab = ImageCms.applyTransform(i, t)
    assert i_lab is not None
    assert i_lab.mode == "LAB"

    k = i_lab.getpixel((0, 0))
    # not a linear luminance map. so L != 128:
    assert k == (137, 128, 128)

    l_data = i_lab.get_flattened_data(0)
    a_data = i_lab.get_flattened_data(1)
    b_data = i_lab.get_flattened_data(2)

    assert l_data == (137.0,) * 100
    assert a_data == (128.0,) * 100
    assert b_data == (128.0,) * 100


def test_lab_color() -> None:
    psRGB = ImageCms.createProfile("sRGB")
    pLab = ImageCms.createProfile("LAB")
    t = ImageCms.buildTransform(psRGB, pLab, "RGB", "LAB")

    # Need to add a type mapping for some PIL type to TYPE_Lab_8 in findLCMSType, and
    # have that mapping work back to a PIL mode (likely RGB).
    i = ImageCms.applyTransform(hopper(), t)
    assert i is not None
    assert_image(i, "LAB", (128, 128))

    # i.save('temp.lab.tif')  # visually verified vs PS.

    assert_image_similar_tofile(i, "Tests/images/hopper.Lab.tif", 3.5)


def test_lab_srgb() -> None:
    psRGB = ImageCms.createProfile("sRGB")
    pLab = ImageCms.createProfile("LAB")
    t = ImageCms.buildTransform(pLab, psRGB, "LAB", "RGB")

    with Image.open("Tests/images/hopper.Lab.tif") as img:
        img_srgb = ImageCms.applyTransform(img, t)
    assert img_srgb is not None

    # img_srgb.save('temp.srgb.tif') # visually verified vs ps.

    assert_image_similar(hopper(), img_srgb, 30)
    assert img_srgb.info["icc_profile"]

    profile = ImageCmsProfile(BytesIO(img_srgb.info["icc_profile"]))
    assert "sRGB" in ImageCms.getProfileDescription(profile)


def test_lab_roundtrip() -> None:
    # check to see if we're at least internally consistent.
    psRGB = ImageCms.createProfile("sRGB")
    pLab = ImageCms.createProfile("LAB")
    t = ImageCms.buildTransform(psRGB, pLab, "RGB", "LAB")

    t2 = ImageCms.buildTransform(pLab, psRGB, "LAB", "RGB")

    i = ImageCms.applyTransform(hopper(), t)
    assert i is not None
    assert i.info["icc_profile"] == ImageCmsProfile(pLab).tobytes()

    out = ImageCms.applyTransform(i, t2)
    assert out is not None
    assert_image_similar(hopper(), out, 2)


def test_profile_tobytes() -> None:
    with Image.open("Tests/images/rgb.jpg") as i:
        p = ImageCms.getOpenProfile(BytesIO(i.info["icc_profile"]))

    p2 = ImageCms.getOpenProfile(BytesIO(p.tobytes()))

    # not the same bytes as the original icc_profile, but it does roundtrip
    assert p.tobytes() == p2.tobytes()
    assert ImageCms.getProfileName(p) == ImageCms.getProfileName(p2)
    assert ImageCms.getProfileDescription(p) == ImageCms.getProfileDescription(p2)


def test_extended_information() -> None:
    skip_missing()
    o = ImageCms.getOpenProfile(SRGB)
    p = o.profile

    def assert_truncated_tuple_equal(
        tup1: tuple[Any, ...] | None, tup2: tuple[Any, ...], digits: int = 10
    ) -> None:
        # Helper function to reduce precision of tuples of floats
        # recursively and then check equality.
        power = 10**digits

        def truncate_tuple(tuple_value: tuple[Any, ...]) -> tuple[Any, ...]:
            return tuple(
                (
                    truncate_tuple(val)
                    if isinstance(val, tuple)
                    else int(val * power) / power
                )
                for val in tuple_value
            )

        assert tup1 is not None
        assert truncate_tuple(tup1) == truncate_tuple(tup2)

    assert p.attributes == 4294967296
    assert_truncated_tuple_equal(
        p.blue_colorant,
        (
            (0.14306640625, 0.06060791015625, 0.7140960693359375),
            (0.1558847490315394, 0.06603820639433387, 0.06060791015625),
        ),
    )
    assert_truncated_tuple_equal(
        p.blue_primary,
        (
            (0.14306641366715667, 0.06060790921083026, 0.7140960805782015),
            (0.15588475410450106, 0.06603820408959558, 0.06060790921083026),
        ),
    )
    assert_truncated_tuple_equal(
        p.chromatic_adaptation,
        (
            (
                (1.04791259765625, 0.0229339599609375, -0.050201416015625),
                (0.02960205078125, 0.9904632568359375, -0.0170745849609375),
                (-0.009246826171875, 0.0150604248046875, 0.7517852783203125),
            ),
            (
                (1.0267159024652783, 0.022470062342089134, 0.0229339599609375),
                (0.02951378324103937, 0.9875098886387147, 0.9904632568359375),
                (-0.012205438066465256, 0.01987915407854985, 0.0150604248046875),
            ),
        ),
    )
    assert p.chromaticity is None
    assert p.clut == {
        0: (False, False, True),
        1: (False, False, True),
        2: (False, False, True),
        3: (False, False, True),
    }

    assert p.colorant_table is None
    assert p.colorant_table_out is None
    assert p.colorimetric_intent is None
    assert p.connection_space == "XYZ "
    assert p.copyright == "Copyright International Color Consortium, 2009"
    assert p.creation_date == datetime.datetime(2009, 2, 27, 21, 36, 31)
    assert p.device_class == "mntr"
    assert_truncated_tuple_equal(
        p.green_colorant,
        (
            (0.3851470947265625, 0.7168731689453125, 0.097076416015625),
            (0.32119769927720654, 0.5978443449048152, 0.7168731689453125),
        ),
    )
    assert_truncated_tuple_equal(
        p.green_primary,
        (
            (0.3851470888162112, 0.7168731974161346, 0.09707641738998518),
            (0.32119768793686687, 0.5978443567149709, 0.7168731974161346),
        ),
    )
    assert p.header_flags == 0
    assert p.header_manufacturer == "\x00\x00\x00\x00"
    assert p.header_model == "\x00\x00\x00\x00"
    assert p.icc_measurement_condition == {
        "backing": (0.0, 0.0, 0.0),
        "flare": 0.0,
        "geo": "unknown",
        "observer": 1,
        "illuminant_type": "D65",
    }
    assert p.icc_version == 33554432
    assert p.icc_viewing_condition is None
    assert p.intent_supported == {
        0: (True, True, True),
        1: (True, True, True),
        2: (True, True, True),
        3: (True, True, True),
    }
    assert p.is_matrix_shaper
    assert p.luminance == ((0.0, 80.0, 0.0), (0.0, 1.0, 80.0))
    assert p.manufacturer is None
    assert_truncated_tuple_equal(
        p.media_black_point,
        (
            (0.012054443359375, 0.0124969482421875, 0.01031494140625),
            (0.34573304157549234, 0.35842450765864337, 0.0124969482421875),
        ),
    )
    assert_truncated_tuple_equal(
        p.media_white_point,
        (
            (0.964202880859375, 1.0, 0.8249053955078125),
            (0.3457029219802284, 0.3585375327567059, 1.0),
        ),
    )
    assert_truncated_tuple_equal(
        (p.media_white_point_temperature,), (5000.722328847392,)
    )
    assert p.model == "IEC 61966-2-1 Default RGB Colour Space - sRGB"

    assert p.perceptual_rendering_intent_gamut is None

    assert p.profile_description == "sRGB IEC61966-2-1 black scaled"
    assert p.profile_id == b")\xf8=\xde\xaf\xf2U\xaexB\xfa\xe4\xca\x839\r"
    assert_truncated_tuple_equal(
        p.red_colorant,
        (
            (0.436065673828125, 0.2224884033203125, 0.013916015625),
            (0.6484536316398539, 0.3308524880306778, 0.2224884033203125),
        ),
    )
    assert_truncated_tuple_equal(
        p.red_primary,
        (
            (0.43606566581047446, 0.22248840582960838, 0.013916015621759925),
            (0.6484536250319214, 0.3308524944738204, 0.22248840582960838),
        ),
    )
    assert p.rendering_intent == 0
    assert p.saturation_rendering_intent_gamut is None
    assert p.screening_description is None
    assert p.target is None
    assert p.technology == "CRT "
    assert p.version == 2.0
    assert p.viewing_condition == "Reference Viewing Condition in IEC 61966-2-1"
    assert p.xcolor_space == "RGB "


def test_non_ascii_path(tmp_path: Path) -> None:
    skip_missing()
    tempfile = str(tmp_path / ("temp_" + chr(128) + ".icc"))
    try:
        shutil.copy(SRGB, tempfile)
    except UnicodeEncodeError:
        pytest.skip("Non-ASCII path could not be created")

    o = ImageCms.getOpenProfile(tempfile)
    p = o.profile
    assert p.model == "IEC 61966-2-1 Default RGB Colour Space - sRGB"


def test_profile_typesafety() -> None:
    # does not segfault
    with pytest.raises(TypeError, match="Invalid type for Profile"):
        ImageCms.ImageCmsProfile(0)  # type: ignore[arg-type]
    with pytest.raises(TypeError, match="Invalid type for Profile"):
        ImageCms.ImageCmsProfile(1)  # type: ignore[arg-type]

    # also check core function
    with pytest.raises(TypeError):
        ImageCms.core.profile_tobytes(0)  # type: ignore[arg-type]
    with pytest.raises(TypeError):
        ImageCms.core.profile_tobytes(1)  # type: ignore[arg-type]

    if not is_pypy():
        # core profile should not be directly instantiable
        with pytest.raises(TypeError):
            ImageCms.core.CmsProfile()
        with pytest.raises(TypeError):
            ImageCms.core.CmsProfile(0)  # type: ignore[call-arg]


@pytest.mark.skipif(is_pypy(), reason="fails on PyPy")
def test_transform_typesafety() -> None:
    # core transform should not be directly instantiable
    with pytest.raises(TypeError):
        ImageCms.core.CmsTransform()
    with pytest.raises(TypeError):
        ImageCms.core.CmsTransform(0)  # type: ignore[call-arg]


def assert_aux_channel_preserved(
    mode: str, transform_in_place: bool, preserved_channel: str
) -> None:
    def create_test_image() -> Image.Image:
        # set up test image with something interesting in the tested aux channel.
        # fmt: off
        nine_grid_deltas = [
            (-1, -1), (-1, 0), (-1, 1),
            (0,  -1),  (0, 0),  (0, 1),
            (1,  -1),  (1, 0),  (1, 1),
        ]
        # fmt: on
        chans = []
        bands = ImageMode.getmode(mode).bands
        for band_ndx in range(len(bands)):
            channel_type = "L"  # 8-bit unorm
            channel_pattern = hopper(channel_type)

            # paste pattern with varying offsets to avoid correlation
            # potentially hiding some bugs (like channels getting mixed).
            paste_offset = (
                int(band_ndx / len(bands) * channel_pattern.size[0]),
                int(band_ndx / (len(bands) * 2) * channel_pattern.size[1]),
            )
            channel_data = Image.new(channel_type, channel_pattern.size)
            for delta in nine_grid_deltas:
                channel_data.paste(
                    channel_pattern,
                    (
                        paste_offset[0] + delta[0] * channel_pattern.size[0],
                        paste_offset[1] + delta[1] * channel_pattern.size[1],
                    ),
                )
            chans.append(channel_data)
        return Image.merge(mode, chans)

    source_image = create_test_image()
    source_image_aux = source_image.getchannel(preserved_channel)

    # create some transform, it doesn't matter which one
    source_profile = ImageCms.createProfile("sRGB")
    destination_profile = ImageCms.createProfile("sRGB")
    t = ImageCms.buildTransform(
        source_profile, destination_profile, inMode=mode, outMode=mode
    )

    # apply transform
    result_image: Image.Image | None
    if transform_in_place:
        ImageCms.applyTransform(source_image, t, inPlace=True)
        result_image = source_image
    else:
        result_image = ImageCms.applyTransform(source_image, t, inPlace=False)
    assert result_image is not None
    result_image_aux = result_image.getchannel(preserved_channel)

    assert_image_equal(source_image_aux, result_image_aux)


def test_preserve_auxiliary_channels_rgba() -> None:
    assert_aux_channel_preserved(
        mode="RGBA", transform_in_place=False, preserved_channel="A"
    )


def test_preserve_auxiliary_channels_rgba_in_place() -> None:
    assert_aux_channel_preserved(
        mode="RGBA", transform_in_place=True, preserved_channel="A"
    )


def test_preserve_auxiliary_channels_rgbx() -> None:
    assert_aux_channel_preserved(
        mode="RGBX", transform_in_place=False, preserved_channel="X"
    )


def test_preserve_auxiliary_channels_rgbx_in_place() -> None:
    assert_aux_channel_preserved(
        mode="RGBX", transform_in_place=True, preserved_channel="X"
    )


def test_auxiliary_channels_isolated() -> None:
    # test data in aux channels does not affect non-aux channels
    aux_channel_formats = [
        # format, profile, color-only format, source test image
        ("RGBA", "sRGB", "RGB", hopper("RGBA")),
        ("RGBX", "sRGB", "RGB", hopper("RGBX")),
        ("LAB", "LAB", "LAB", Image.open("Tests/images/hopper.Lab.tif")),
    ]
    for src_format in aux_channel_formats:
        for dst_format in aux_channel_formats:
            for transform_in_place in [True, False]:
                # inplace only if format doesn't change
                if transform_in_place and src_format[0] != dst_format[0]:
                    continue

                # convert with and without AUX data, test colors are equal
                src_colorSpace = cast(Literal["LAB", "XYZ", "sRGB"], src_format[1])
                source_profile = ImageCms.createProfile(src_colorSpace)
                dst_colorSpace = cast(Literal["LAB", "XYZ", "sRGB"], dst_format[1])
                destination_profile = ImageCms.createProfile(dst_colorSpace)
                source_image = src_format[3]
                test_transform = ImageCms.buildTransform(
                    source_profile,
                    destination_profile,
                    inMode=src_format[0],
                    outMode=dst_format[0],
                )

                # test conversion from aux-ful source
                test_image: Image.Image | None
                if transform_in_place:
                    test_image = source_image.copy()
                    ImageCms.applyTransform(test_image, test_transform, inPlace=True)
                else:
                    test_image = ImageCms.applyTransform(
                        source_image, test_transform, inPlace=False
                    )
                assert test_image is not None

                # reference conversion from aux-less source
                reference_transform = ImageCms.buildTransform(
                    source_profile,
                    destination_profile,
                    inMode=src_format[2],
                    outMode=dst_format[2],
                )
                reference_image = ImageCms.applyTransform(
                    source_image.convert(src_format[2]), reference_transform
                )
                assert reference_image is not None
                assert_image_equal(test_image.convert(dst_format[2]), reference_image)


@pytest.mark.parametrize("mode", ("RGB", "RGBA", "RGBX"))
def test_rgb_lab(mode: str) -> None:
    im = Image.new(mode, (1, 1))
    converted_im = im.convert("LAB")
    assert converted_im.getpixel((0, 0)) == (0, 128, 128)

    im = Image.new("LAB", (1, 1), (255, 0, 0))
    converted_im = im.convert(mode)
    value = converted_im.getpixel((0, 0))
    assert isinstance(value, tuple)
    assert value[:3] == (0, 255, 255)


def test_cmyk_lab() -> None:
    im = Image.new("CMYK", (1, 1))
    converted_im = im.convert("LAB")
    assert converted_im.getpixel((0, 0)) == (255, 128, 128)


def test_deprecation() -> None:
    profile = ImageCmsProfile(ImageCms.createProfile("sRGB"))
    with pytest.warns(
        DeprecationWarning, match="ImageCms.ImageCmsProfile.product_name"
    ):
        profile.product_name
    with pytest.warns(
        DeprecationWarning, match="ImageCms.ImageCmsProfile.product_info"
    ):
        profile.product_info
    with pytest.raises(AttributeError):
        profile.this_attribute_does_not_exist



from __future__ import annotations

import pytest

from PIL import Image

from .helper import hopper


@pytest.mark.parametrize(
    "mode, dest_modes",
    (
        ("L", ["I", "F", "LA", "RGB", "RGBA", "RGBX", "CMYK", "YCbCr", "HSV"]),
        ("I", ["L", "F"]),  # Technically I;32 can work for any 4x8bit storage.
        ("F", ["I", "L", "LA", "RGB", "RGBA", "RGBX", "CMYK", "YCbCr", "HSV"]),
        ("LA", ["L", "F"]),
        ("RGB", ["L", "F"]),
        ("RGBA", ["L", "F"]),
        ("RGBX", ["L", "F"]),
        ("CMYK", ["L", "F"]),
        ("YCbCr", ["L", "F"]),
        ("HSV", ["L", "F"]),
    ),
)
def test_invalid_array_type(mode: str, dest_modes: list[str]) -> None:
    img = hopper(mode)
    for dest_mode in dest_modes:
        with pytest.raises(ValueError):
            Image.fromarrow(img, dest_mode, img.size)


def test_invalid_array_size() -> None:
    img = hopper("RGB")

    assert img.size != (10, 10)
    with pytest.raises(ValueError):
        Image.fromarrow(img, "RGB", (10, 10))


def test_release_schema() -> None:
    # these should not error out, valgrind should be clean
    img = hopper("L")
    schema = img.__arrow_c_schema__()
    del schema


def test_release_array() -> None:
    # these should not error out, valgrind should be clean
    img = hopper("L")
    array, schema = img.__arrow_c_array__()
    del array
    del schema


def test_readonly() -> None:
    img = hopper("L")
    reloaded = Image.fromarrow(img, img.mode, img.size)
    assert reloaded.readonly == 1
    reloaded._readonly = 0
    assert reloaded.readonly == 1


def test_multiblock_l_image() -> None:
    block_size = Image.core.get_block_size()

    # check a 2 block image in single channel mode
    size = (4096, 2 * block_size // 4096)
    img = Image.new("L", size, 128)

    with pytest.raises(ValueError):
        schema, arr = img.__arrow_c_array__()


def test_multiblock_rgba_image() -> None:
    block_size = Image.core.get_block_size()

    # check a 2 block image in 4 channel mode
    size = (4096, (block_size // 4096) // 2)
    img = Image.new("RGBA", size, (128, 127, 126, 125))

    with pytest.raises(ValueError):
        schema, arr = img.__arrow_c_array__()


def test_multiblock_l_schema() -> None:
    block_size = Image.core.get_block_size()

    # check a 2 block image in single channel mode
    size = (4096, 2 * block_size // 4096)
    img = Image.new("L", size, 128)

    with pytest.raises(ValueError):
        img.__arrow_c_schema__()


def test_multiblock_rgba_schema() -> None:
    block_size = Image.core.get_block_size()

    # check a 2 block image in 4 channel mode
    size = (4096, (block_size // 4096) // 2)
    img = Image.new("RGBA", size, (128, 127, 126, 125))

    with pytest.raises(ValueError):
        img.__arrow_c_schema__()


def test_singleblock_l_image() -> None:
    Image.core.set_use_block_allocator(1)

    block_size = Image.core.get_block_size()

    # check a 2 block image in 4 channel mode
    size = (4096, 2 * (block_size // 4096))
    img = Image.new("L", size, 128)
    assert img.im.isblock()

    schema, arr = img.__arrow_c_array__()
    assert schema
    assert arr

    Image.core.set_use_block_allocator(0)


def test_singleblock_rgba_image() -> None:
    Image.core.set_use_block_allocator(1)
    block_size = Image.core.get_block_size()

    # check a 2 block image in 4 channel mode
    size = (4096, (block_size // 4096) // 2)
    img = Image.new("RGBA", size, (128, 127, 126, 125))
    assert img.im.isblock()

    schema, arr = img.__arrow_c_array__()
    assert schema
    assert arr
    Image.core.set_use_block_allocator(0)


def test_singleblock_l_schema() -> None:
    Image.core.set_use_block_allocator(1)
    block_size = Image.core.get_block_size()

    # check a 2 block image in single channel mode
    size = (4096, 2 * block_size // 4096)
    img = Image.new("L", size, 128)
    assert img.im.isblock()

    schema = img.__arrow_c_schema__()
    assert schema
    Image.core.set_use_block_allocator(0)


def test_singleblock_rgba_schema() -> None:
    Image.core.set_use_block_allocator(1)
    block_size = Image.core.get_block_size()

    # check a 2 block image in 4 channel mode
    size = (4096, (block_size // 4096) // 2)
    img = Image.new("RGBA", size, (128, 127, 126, 125))
    assert img.im.isblock()

    schema = img.__arrow_c_schema__()
    assert schema
    Image.core.set_use_block_allocator(0)



from __future__ import annotations

from pathlib import Path

import pytest

from PIL import Image, QoiImagePlugin

from .helper import assert_image_equal_tofile, hopper


def test_sanity() -> None:
    with Image.open("Tests/images/hopper.qoi") as im:
        assert im.mode == "RGB"
        assert im.size == (128, 128)
        assert im.format == "QOI"

        assert_image_equal_tofile(im, "Tests/images/hopper.png")

    with Image.open("Tests/images/pil123rgba.qoi") as im:
        assert im.mode == "RGBA"
        assert im.size == (162, 150)
        assert im.format == "QOI"

        assert_image_equal_tofile(im, "Tests/images/pil123rgba.png")


def test_invalid_file() -> None:
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        QoiImagePlugin.QoiImageFile(invalid_file)


def test_op_index() -> None:
    # QOI_OP_INDEX as the first chunk
    with Image.open("Tests/images/op_index.qoi") as im:
        assert im.getpixel((0, 0)) == (0, 0, 0, 0)


def test_save(tmp_path: Path) -> None:
    f = tmp_path / "temp.qoi"

    im = hopper()
    im.save(f, colorspace="sRGB")

    assert_image_equal_tofile(im, f)

    for path in ("Tests/images/default_font.png", "Tests/images/pil123rgba.png"):
        with Image.open(path) as im:
            im.save(f)

            assert_image_equal_tofile(im, f)

    im = hopper("P")
    with pytest.raises(ValueError, match="Unsupported QOI image mode"):
        im.save(f)



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import FitsImagePlugin, Image

from .helper import assert_image_equal, assert_image_equal_tofile, hopper

TEST_FILE = "Tests/images/hopper.fits"


def test_open() -> None:
    # Act
    with Image.open(TEST_FILE) as im:
        # Assert
        assert im.format == "FITS"
        assert im.size == (128, 128)
        assert im.mode == "L"

        assert_image_equal(im, hopper("L"))


def test_gzip1() -> None:
    with Image.open("Tests/images/m13_gzip.fits") as im:
        assert_image_equal_tofile(im, "Tests/images/m13.fits")


def test_invalid_file() -> None:
    # Arrange
    invalid_file = "Tests/images/flower.jpg"

    # Act / Assert
    with pytest.raises(SyntaxError):
        FitsImagePlugin.FitsImageFile(invalid_file)


def test_truncated_fits() -> None:
    # No END to headers
    image_data = b"SIMPLE  =                    T" + b" " * 50 + b"TRUNCATE"
    with pytest.raises(OSError):
        FitsImagePlugin.FitsImageFile(BytesIO(image_data))


def test_naxis_zero() -> None:
    # This test image has been manually hexedited
    # to set the number of data axes to zero
    with pytest.raises(ValueError):
        with Image.open("Tests/images/hopper_naxis_zero.fits"):
            pass


def test_comment() -> None:
    image_data = b"SIMPLE  =                    T / comment string"
    with pytest.raises(OSError):
        FitsImagePlugin.FitsImageFile(BytesIO(image_data))



from __future__ import annotations

from collections.abc import Generator
from pathlib import Path

import pytest

from PIL import GifImagePlugin, Image, WebPImagePlugin

from .helper import (
    assert_image_equal,
    assert_image_similar,
    has_feature_version,
    is_big_endian,
    skip_unless_feature,
)

pytestmark = skip_unless_feature("webp")


def test_n_frames() -> None:
    """Ensure that WebP format sets n_frames and is_animated attributes correctly."""

    with Image.open("Tests/images/hopper.webp") as im:
        assert isinstance(im, WebPImagePlugin.WebPImageFile)
        assert im.n_frames == 1
        assert not im.is_animated

    with Image.open("Tests/images/iss634.webp") as im:
        assert isinstance(im, WebPImagePlugin.WebPImageFile)
        assert im.n_frames == 42
        assert im.is_animated


def test_write_animation_L(tmp_path: Path) -> None:
    """
    Convert an animated GIF to animated WebP, then compare the frame count, and first
    and last frames to ensure they're visually similar.
    """

    with Image.open("Tests/images/iss634.gif") as orig:
        assert isinstance(orig, GifImagePlugin.GifImageFile)
        assert orig.n_frames > 1

        temp_file = tmp_path / "temp.webp"
        orig.save(temp_file, save_all=True)
        with Image.open(temp_file) as im:
            assert isinstance(im, WebPImagePlugin.WebPImageFile)
            assert im.n_frames == orig.n_frames

            # Compare first and last frames to the original animated GIF
            orig.load()
            im.load()
            assert_image_similar(im, orig.convert("RGBA"), 32.9)

            if is_big_endian() and not has_feature_version("webp", "1.2.2"):
                pytest.skip("Fails with libwebp earlier than 1.2.2")
            orig.seek(orig.n_frames - 1)
            im.seek(im.n_frames - 1)
            orig.load()
            im.load()
            assert_image_similar(im, orig.convert("RGBA"), 32.9)


def test_write_animation_RGB(tmp_path: Path) -> None:
    """
    Write an animated WebP from RGB frames, and ensure the frames
    are visually similar to the originals.
    """

    def check(temp_file: Path) -> None:
        with Image.open(temp_file) as im:
            assert isinstance(im, WebPImagePlugin.WebPImageFile)
            assert im.n_frames == 2

            # Compare first frame to original
            im.load()
            assert_image_equal(im, frame1.convert("RGBA"))

            # Compare second frame to original
            if is_big_endian() and not has_feature_version("webp", "1.2.2"):
                pytest.skip("Fails with libwebp earlier than 1.2.2")
            im.seek(1)
            im.load()
            assert_image_equal(im, frame2.convert("RGBA"))

    with Image.open("Tests/images/anim_frame1.webp") as frame1:
        with Image.open("Tests/images/anim_frame2.webp") as frame2:
            temp_file1 = tmp_path / "temp.webp"
            frame1.copy().save(
                temp_file1, save_all=True, append_images=[frame2], lossless=True
            )
            check(temp_file1)

            # Tests appending using a generator
            def im_generator(
                ims: list[Image.Image],
            ) -> Generator[Image.Image, None, None]:
                yield from ims

            temp_file2 = tmp_path / "temp_generator.webp"
            frame1.copy().save(
                temp_file2,
                save_all=True,
                append_images=im_generator([frame2]),
                lossless=True,
            )
            check(temp_file2)


def test_timestamp_and_duration(tmp_path: Path) -> None:
    """
    Try passing a list of durations, and make sure the encoded
    timestamps and durations are correct.
    """

    durations = [0, 10, 20, 30, 40]
    temp_file = tmp_path / "temp.webp"
    with Image.open("Tests/images/anim_frame1.webp") as frame1:
        with Image.open("Tests/images/anim_frame2.webp") as frame2:
            frame1.save(
                temp_file,
                save_all=True,
                append_images=[frame2, frame1, frame2, frame1],
                duration=durations,
            )

    with Image.open(temp_file) as im:
        assert isinstance(im, WebPImagePlugin.WebPImageFile)
        assert im.n_frames == 5
        assert im.is_animated

        # Check that timestamps and durations match original values specified
        ts = 0
        for frame in range(im.n_frames):
            im.seek(frame)
            im.load()
            assert im.info["duration"] == durations[frame]
            assert im.info["timestamp"] == ts
            ts += durations[frame]


def test_float_duration(tmp_path: Path) -> None:
    temp_file = tmp_path / "temp.webp"
    with Image.open("Tests/images/iss634.apng") as im:
        assert im.info["duration"] == 70.0

        im.save(temp_file, save_all=True)

    with Image.open(temp_file) as reloaded:
        reloaded.load()
        assert reloaded.info["duration"] == 70


def test_seeking(tmp_path: Path) -> None:
    """
    Create an animated WebP file, and then try seeking through frames in reverse-order,
    verifying the timestamps and durations are correct.
    """

    dur = 33
    temp_file = tmp_path / "temp.webp"
    with Image.open("Tests/images/anim_frame1.webp") as frame1:
        with Image.open("Tests/images/anim_frame2.webp") as frame2:
            frame1.save(
                temp_file,
                save_all=True,
                append_images=[frame2, frame1, frame2, frame1],
                duration=dur,
            )

    with Image.open(temp_file) as im:
        assert isinstance(im, WebPImagePlugin.WebPImageFile)
        assert im.n_frames == 5
        assert im.is_animated

        # Traverse frames in reverse, checking timestamps and durations
        ts = dur * (im.n_frames - 1)
        for frame in reversed(range(im.n_frames)):
            im.seek(frame)
            im.load()
            assert im.info["duration"] == dur
            assert im.info["timestamp"] == ts
            ts -= dur


def test_seek_errors() -> None:
    with Image.open("Tests/images/iss634.webp") as im:
        with pytest.raises(EOFError):
            im.seek(-1)

        with pytest.raises(EOFError):
            im.seek(42)


def test_alpha_quality(tmp_path: Path) -> None:
    with Image.open("Tests/images/transparent.png") as im:
        first_frame = Image.new("L", im.size)

        out = tmp_path / "temp.webp"
        first_frame.save(out, save_all=True, append_images=[im])

        out_quality = tmp_path / "quality.webp"
        first_frame.save(
            out_quality, save_all=True, append_images=[im], alpha_quality=50
        )
        with Image.open(out) as reloaded:
            reloaded.seek(1)
            with Image.open(out_quality) as reloaded_quality:
                reloaded_quality.seek(1)
                assert reloaded.tobytes() != reloaded_quality.tobytes()



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import Image

from .helper import assert_image_equal


def test_load_raw() -> None:
    with Image.open("Tests/images/hopper.pcd") as im:
        assert im.size == (768, 512)
        im.load()  # should not segfault.

    # Note that this image was created with a resized hopper
    # image, which was then converted to pcd with imagemagick
    # and the colors are wonky in Pillow.  It's unclear if this
    # is a pillow or a convert issue, as other images not generated
    # from convert look find on pillow and not imagemagick.

    # target = hopper().resize((768,512))
    # assert_image_similar(im, target, 10)


@pytest.mark.parametrize("orientation", (1, 3))
def test_rotated(orientation: int) -> None:
    with open("Tests/images/hopper.pcd", "rb") as fp:
        data = bytearray(fp.read())
    data[2048 + 1538] = orientation
    f = BytesIO(data)
    with Image.open(f) as im:
        assert im.size == (512, 768)

        with Image.open("Tests/images/hopper.pcd") as expected:
            assert_image_equal(
                im, expected.rotate(90 if orientation == 1 else 270, expand=True)
            )



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import Image, XpmImagePlugin

from .helper import assert_image_similar, hopper

TEST_FILE = "Tests/images/hopper.xpm"


def test_sanity() -> None:
    with Image.open(TEST_FILE) as im:
        im.load()
        assert im.mode == "P"
        assert im.size == (128, 128)
        assert im.format == "XPM"

        # large error due to quantization->44 colors.
        assert_image_similar(im.convert("RGB"), hopper(), 23)


def test_bpp2() -> None:
    with Image.open("Tests/images/hopper_bpp2.xpm") as im:
        assert_image_similar(im.convert("RGB"), hopper(), 11)


def test_rgb() -> None:
    with Image.open("Tests/images/hopper_rgb.xpm") as im:
        assert im.mode == "RGB"
        assert_image_similar(im, hopper(), 16)


def test_truncated_header() -> None:
    data = b"/* XPM */"
    with pytest.raises(SyntaxError, match="broken XPM file"):
        with XpmImagePlugin.XpmImageFile(BytesIO(data)):
            pass


def test_cannot_read_color() -> None:
    with open(TEST_FILE, "rb") as fp:
        data = fp.read().split(b"#")[0]
    with pytest.raises(ValueError, match="cannot read this XPM file"):
        with Image.open(BytesIO(data)):
            pass

    with pytest.raises(ValueError, match="cannot read this XPM file"):
        with Image.open(BytesIO(data + b"invalid")):
            pass


def test_not_enough_image_data() -> None:
    with open(TEST_FILE, "rb") as fp:
        data = fp.read().split(b"/* pixels */")[0]
    with Image.open(BytesIO(data)) as im:
        with pytest.raises(ValueError, match="not enough image data"):
            im.load()


def test_invalid_file() -> None:
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        XpmImagePlugin.XpmImageFile(invalid_file)


def test_load_read() -> None:
    # Arrange
    with Image.open(TEST_FILE) as im:
        assert isinstance(im, XpmImagePlugin.XpmImageFile)
        dummy_bytes = 1

        # Act
        data = im.load_read(dummy_bytes)

    # Assert
    assert len(data) == 16384



from __future__ import annotations

import gc
import os
import re
import warnings
from collections.abc import Generator, Sequence
from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from typing import Any

import pytest

from PIL import (
    AvifImagePlugin,
    GifImagePlugin,
    Image,
    ImageDraw,
    ImageFile,
    UnidentifiedImageError,
    features,
)

from .helper import (
    PillowLeakTestCase,
    assert_image,
    assert_image_similar,
    assert_image_similar_tofile,
    hopper,
    skip_unless_feature,
    skip_unless_feature_version,
)

try:
    from PIL import _avif

    HAVE_AVIF = True
except ImportError:
    HAVE_AVIF = False


TEST_AVIF_FILE = "Tests/images/avif/hopper.avif"


def assert_xmp_orientation(xmp: bytes, expected: int) -> None:
    assert int(xmp.split(b'tiff:Orientation="')[1].split(b'"')[0]) == expected


def roundtrip(im: Image.Image, **options: Any) -> ImageFile.ImageFile:
    out = BytesIO()
    im.save(out, "AVIF", **options)
    return Image.open(out)


def skip_unless_avif_decoder(codec_name: str) -> pytest.MarkDecorator:
    reason = f"{codec_name} decode not available"
    return pytest.mark.skipif(
        not HAVE_AVIF or not _avif.decoder_codec_available(codec_name), reason=reason
    )


def skip_unless_avif_encoder(codec_name: str) -> pytest.MarkDecorator:
    reason = f"{codec_name} encode not available"
    return pytest.mark.skipif(
        not HAVE_AVIF or not _avif.encoder_codec_available(codec_name), reason=reason
    )


def is_docker_qemu() -> bool:
    try:
        init_proc_exe = os.readlink("/proc/1/exe")
    except (FileNotFoundError, PermissionError):
        return False
    return "qemu" in init_proc_exe


class TestUnsupportedAvif:
    def test_unsupported(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(AvifImagePlugin, "SUPPORTED", False)

        with pytest.raises(UnidentifiedImageError):
            with pytest.warns(UserWarning, match="AVIF support not installed"):
                with Image.open(TEST_AVIF_FILE):
                    pass

    def test_unsupported_open(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(AvifImagePlugin, "SUPPORTED", False)

        with pytest.raises(SyntaxError):
            AvifImagePlugin.AvifImageFile(TEST_AVIF_FILE)


@skip_unless_feature("avif")
class TestFileAvif:
    def test_version(self) -> None:
        version = features.version_module("avif")
        assert version is not None
        assert re.search(r"^\d+\.\d+\.\d+$", version)

    def test_codec_version(self) -> None:
        assert AvifImagePlugin.get_codec_version("unknown") is None

        for codec_name in ("aom", "dav1d", "rav1e", "svt"):
            codec_version = AvifImagePlugin.get_codec_version(codec_name)
            if _avif.decoder_codec_available(
                codec_name
            ) or _avif.encoder_codec_available(codec_name):
                assert codec_version is not None
                assert re.search(r"^v?\d+\.\d+\.\d+(-([a-z\d])+)*$", codec_version)
            else:
                assert codec_version is None

    def test_read(self) -> None:
        """
        Can we read an AVIF file without error?
        Does it have the bits we expect?
        """

        with Image.open(TEST_AVIF_FILE) as image:
            assert image.mode == "RGB"
            assert image.size == (128, 128)
            assert image.format == "AVIF"
            assert image.get_format_mimetype() == "image/avif"

            # generated with:
            # avifdec hopper.avif hopper_avif_write.png
            assert_image_similar_tofile(
                image, "Tests/images/avif/hopper_avif_write.png", 11.5
            )

    @skip_unless_feature_version("avif", "1.3.0")
    def test_write_l(self) -> None:
        im = hopper("L")
        reloaded = roundtrip(im)

        assert reloaded.mode == "L"
        assert_image_similar(reloaded, im, 1.69)

    def test_write_rgb(self, tmp_path: Path) -> None:
        """
        Can we write a RGB mode file to avif without error?
        Does it have the bits we expect?
        """

        temp_file = tmp_path / "temp.avif"

        im = hopper()
        im.save(temp_file)
        with Image.open(temp_file) as reloaded:
            assert reloaded.mode == "RGB"
            assert reloaded.size == (128, 128)
            assert reloaded.format == "AVIF"

            # avifdec hopper.avif avif/hopper_avif_write.png
            assert_image_similar_tofile(
                reloaded, "Tests/images/avif/hopper_avif_write.png", 7.0
            )

            # This test asserts that the images are similar. If the average pixel
            # difference between the two images is less than the epsilon value,
            # then we're going to accept that it's a reasonable lossy version of
            # the image.
            assert_image_similar(reloaded, im, 9.5)

    def test_AvifEncoder_with_invalid_args(self) -> None:
        """
        Calling encoder functions with no arguments should result in an error.
        """
        with pytest.raises(TypeError):
            _avif.AvifEncoder()

    def test_AvifDecoder_with_invalid_args(self) -> None:
        """
        Calling decoder functions with no arguments should result in an error.
        """
        with pytest.raises(TypeError):
            _avif.AvifDecoder()

    def test_invalid_dimensions(self, tmp_path: Path) -> None:
        test_file = tmp_path / "temp.avif"
        im = Image.new("RGB", (0, 0))
        with pytest.raises(ValueError):
            im.save(test_file)

    def test_encoder_finish_none_error(
        self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
    ) -> None:
        """Save should raise an OSError if AvifEncoder.finish returns None"""

        class _mock_avif:
            class AvifEncoder:
                def __init__(self, *args: Any) -> None:
                    pass

                def add(self, *args: Any) -> None:
                    pass

                def finish(self) -> None:
                    return None

        monkeypatch.setattr(AvifImagePlugin, "_avif", _mock_avif)

        im = Image.new("RGB", (150, 150))
        test_file = tmp_path / "temp.avif"
        with pytest.raises(OSError):
            im.save(test_file)

    def test_no_resource_warning(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            with warnings.catch_warnings():
                warnings.simplefilter("error")

                im.save(tmp_path / "temp.avif")

    @pytest.mark.parametrize("major_brand", [b"avif", b"avis", b"mif1", b"msf1"])
    def test_accept_ftyp_brands(self, major_brand: bytes) -> None:
        data = b"\x00\x00\x00\x1cftyp%s\x00\x00\x00\x00" % major_brand
        assert AvifImagePlugin._accept(data) is True

    def test_file_pointer_could_be_reused(self) -> None:
        with open(TEST_AVIF_FILE, "rb") as blob:
            with Image.open(blob) as im:
                im.load()
            with Image.open(blob) as im:
                im.load()

    def test_background_from_gif(self, tmp_path: Path) -> None:
        with Image.open("Tests/images/chi.gif") as im:
            original_value = im.convert("RGB").getpixel((1, 1))
            assert isinstance(original_value, tuple)

            # Save as AVIF
            out_avif = tmp_path / "temp.avif"
            im.save(out_avif, save_all=True)

        # Save as GIF
        out_gif = tmp_path / "temp.gif"
        with Image.open(out_avif) as im:
            im.save(out_gif)

        with Image.open(out_gif) as reread:
            reread_value = reread.convert("RGB").getpixel((1, 1))
        assert isinstance(reread_value, tuple)
        difference = sum([abs(original_value[i] - reread_value[i]) for i in range(3)])
        assert difference <= 6

    def test_save_single_frame(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.avif"
        with Image.open("Tests/images/chi.gif") as im:
            im.save(temp_file)
        with Image.open(temp_file) as im:
            assert isinstance(im, AvifImagePlugin.AvifImageFile)
            assert im.n_frames == 1

    def test_invalid_file(self) -> None:
        invalid_file = "Tests/images/flower.jpg"

        with pytest.raises(SyntaxError):
            AvifImagePlugin.AvifImageFile(invalid_file)

    def test_load_transparent_rgb(self) -> None:
        test_file = "Tests/images/avif/transparency.avif"
        with Image.open(test_file) as im:
            assert_image(im, "RGBA", (64, 64))

            # image has 876 transparent pixels
            colors = im.getchannel("A").getcolors()
            assert colors is not None
            assert colors[0] == (876, 0)

    def test_save_transparent(self, tmp_path: Path) -> None:
        im = Image.new("RGBA", (10, 10), (0, 0, 0, 0))
        assert im.getcolors() == [(100, (0, 0, 0, 0))]

        test_file = tmp_path / "temp.avif"
        im.save(test_file)

        # check if saved image contains the same transparency
        with Image.open(test_file) as im:
            assert_image(im, "RGBA", (10, 10))
            assert im.getcolors() == [(100, (0, 0, 0, 0))]

    def test_save_icc_profile(self) -> None:
        with Image.open("Tests/images/avif/icc_profile_none.avif") as im:
            assert "icc_profile" not in im.info

            with Image.open("Tests/images/avif/icc_profile.avif") as with_icc:
                expected_icc = with_icc.info["icc_profile"]
                assert expected_icc is not None

                im = roundtrip(im, icc_profile=expected_icc)
                assert im.info["icc_profile"] == expected_icc

    def test_discard_icc_profile(self) -> None:
        with Image.open("Tests/images/avif/icc_profile.avif") as im:
            im = roundtrip(im, icc_profile=None)
        assert "icc_profile" not in im.info

    def test_roundtrip_icc_profile(self) -> None:
        with Image.open("Tests/images/avif/icc_profile.avif") as im:
            expected_icc = im.info["icc_profile"]

            im = roundtrip(im)
        assert im.info["icc_profile"] == expected_icc

    def test_roundtrip_no_icc_profile(self) -> None:
        with Image.open("Tests/images/avif/icc_profile_none.avif") as im:
            assert "icc_profile" not in im.info

            im = roundtrip(im)
        assert "icc_profile" not in im.info

    def test_exif(self) -> None:
        # With an EXIF chunk
        with Image.open("Tests/images/avif/exif.avif") as im:
            exif = im.getexif()
        assert exif[274] == 1

        with Image.open("Tests/images/avif/xmp_tags_orientation.avif") as im:
            exif = im.getexif()
        assert exif[274] == 3

    @pytest.mark.parametrize("use_bytes", [True, False])
    @pytest.mark.parametrize("orientation", [1, 2, 3, 4, 5, 6, 7, 8])
    def test_exif_save(
        self,
        tmp_path: Path,
        use_bytes: bool,
        orientation: int,
    ) -> None:
        exif = Image.Exif()
        exif[274] = orientation
        exif_data = exif.tobytes()
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            im.save(test_file, exif=exif_data if use_bytes else exif)

        with Image.open(test_file) as reloaded:
            if orientation == 1:
                assert "exif" not in reloaded.info
            else:
                assert reloaded.getexif()[274] == orientation
                assert reloaded.info["exif"] == exif_data

    def test_exif_without_orientation(self, tmp_path: Path) -> None:
        exif = Image.Exif()
        exif[272] = b"test"
        exif_data = exif.tobytes()
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            im.save(test_file, exif=exif)

        with Image.open(test_file) as reloaded:
            assert reloaded.info["exif"] == exif_data

    def test_exif_invalid(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(SyntaxError):
                im.save(test_file, exif=b"invalid")

    @pytest.mark.parametrize(
        "rot, mir, exif_orientation",
        [
            (0, 0, 4),
            (0, 1, 2),
            (1, 0, 5),
            (1, 1, 7),
            (2, 0, 2),
            (2, 1, 4),
            (3, 0, 7),
            (3, 1, 5),
        ],
    )
    def test_rot_mir_exif(
        self, rot: int, mir: int, exif_orientation: int, tmp_path: Path
    ) -> None:
        with Image.open(f"Tests/images/avif/rot{rot}mir{mir}.avif") as im:
            exif = im.getexif()
            assert exif[274] == exif_orientation

            test_file = tmp_path / "temp.avif"
            im.save(test_file, exif=exif)
        with Image.open(test_file) as reloaded:
            assert reloaded.getexif()[274] == exif_orientation

    def test_xmp(self) -> None:
        with Image.open("Tests/images/avif/xmp_tags_orientation.avif") as im:
            xmp = im.info["xmp"]
        assert_xmp_orientation(xmp, 3)

    def test_xmp_save(self, tmp_path: Path) -> None:
        xmp_arg = "\n".join(
            [
                '<?xpacket begin="" id="W5M0MpCehiHzreSzNTczkc9d"?>',
                '<x:xmpmeta xmlns:x="adobe:ns:meta/">',
                ' <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">',
                '  <rdf:Description rdf:about=""',
                '    xmlns:tiff="http://ns.adobe.com/tiff/1.0/"',
                '   tiff:Orientation="1"/>',
                " </rdf:RDF>",
                "</x:xmpmeta>",
                '<?xpacket end="r"?>',
            ]
        )
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            im.save(test_file, xmp=xmp_arg)

        with Image.open(test_file) as reloaded:
            xmp = reloaded.info["xmp"]
        assert_xmp_orientation(xmp, 1)

    def test_tell(self) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            assert im.tell() == 0

    def test_seek(self) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            im.seek(0)

            with pytest.raises(EOFError):
                im.seek(1)

    @pytest.mark.parametrize("subsampling", ["4:4:4", "4:2:2", "4:2:0", "4:0:0"])
    def test_encoder_subsampling(self, tmp_path: Path, subsampling: str) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            im.save(test_file, subsampling=subsampling)

    @skip_unless_feature_version("avif", "1.3.0")
    def test_encoding_subsampling_400(self) -> None:
        im = hopper()
        reloaded = roundtrip(im, subsampling="4:0:0")

        assert reloaded.mode == "L"
        assert_image_similar(reloaded, im.convert("L"), 1.69)

    def test_encoder_subsampling_invalid(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(ValueError):
                im.save(test_file, subsampling="foo")

    @pytest.mark.parametrize("value", ["full", "limited"])
    def test_encoder_range(self, tmp_path: Path, value: str) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            im.save(test_file, range=value)

    def test_encoder_range_invalid(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(ValueError):
                im.save(test_file, range="foo")

    @skip_unless_avif_encoder("aom")
    def test_encoder_codec_param(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            im.save(test_file, codec="aom")

    def test_encoder_codec_invalid(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(ValueError):
                im.save(test_file, codec="foo")

    @skip_unless_avif_decoder("dav1d")
    def test_decoder_codec_cannot_encode(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(ValueError):
                im.save(test_file, codec="dav1d")

    @skip_unless_avif_encoder("aom")
    @pytest.mark.parametrize(
        "advanced",
        [
            {"tune": "psnr"},
            (("tune", "psnr"),),
            [("tune", "psnr")],
        ],
    )
    def test_encoder_advanced_codec_options(
        self, advanced: dict[str, str] | Sequence[tuple[str, str]]
    ) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            ctrl_buf = BytesIO()
            im.save(ctrl_buf, "AVIF", codec="aom")
            test_buf = BytesIO()
            im.save(
                test_buf,
                "AVIF",
                codec="aom",
                advanced=advanced,
            )
            assert ctrl_buf.getvalue() != test_buf.getvalue()

    @skip_unless_avif_encoder("aom")
    @pytest.mark.parametrize("advanced", [{"foo": "bar"}, {"foo": 1234}, 1234])
    def test_encoder_advanced_codec_options_invalid(
        self, tmp_path: Path, advanced: dict[str, str] | int
    ) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(ValueError):
                im.save(test_file, codec="aom", advanced=advanced)

    @skip_unless_avif_decoder("aom")
    def test_decoder_codec_param(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(AvifImagePlugin, "DECODE_CODEC_CHOICE", "aom")

        with Image.open(TEST_AVIF_FILE) as im:
            assert im.size == (128, 128)

    @skip_unless_avif_encoder("rav1e")
    def test_encoder_codec_cannot_decode(
        self, monkeypatch: pytest.MonkeyPatch, tmp_path: Path
    ) -> None:
        monkeypatch.setattr(AvifImagePlugin, "DECODE_CODEC_CHOICE", "rav1e")

        with pytest.raises(ValueError):
            with Image.open(TEST_AVIF_FILE):
                pass

    def test_decoder_codec_invalid(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setattr(AvifImagePlugin, "DECODE_CODEC_CHOICE", "foo")

        with pytest.raises(ValueError):
            with Image.open(TEST_AVIF_FILE):
                pass

    @skip_unless_avif_encoder("aom")
    def test_encoder_codec_available(self) -> None:
        assert _avif.encoder_codec_available("aom") is True

    def test_encoder_codec_available_bad_params(self) -> None:
        with pytest.raises(TypeError):
            _avif.encoder_codec_available()

    @skip_unless_avif_decoder("dav1d")
    def test_encoder_codec_available_cannot_decode(self) -> None:
        assert _avif.encoder_codec_available("dav1d") is False

    def test_encoder_codec_available_invalid(self) -> None:
        assert _avif.encoder_codec_available("foo") is False

    def test_encoder_quality_valueerror(self, tmp_path: Path) -> None:
        with Image.open(TEST_AVIF_FILE) as im:
            test_file = tmp_path / "temp.avif"
            with pytest.raises(ValueError):
                im.save(test_file, quality="invalid")

    @skip_unless_avif_decoder("aom")
    def test_decoder_codec_available(self) -> None:
        assert _avif.decoder_codec_available("aom") is True

    def test_decoder_codec_available_bad_params(self) -> None:
        with pytest.raises(TypeError):
            _avif.decoder_codec_available()

    @skip_unless_avif_encoder("rav1e")
    def test_decoder_codec_available_cannot_decode(self) -> None:
        assert _avif.decoder_codec_available("rav1e") is False

    def test_decoder_codec_available_invalid(self) -> None:
        assert _avif.decoder_codec_available("foo") is False

    def test_p_mode_transparency(self, tmp_path: Path) -> None:
        im = Image.new("P", size=(64, 64))
        draw = ImageDraw.Draw(im)
        draw.rectangle(xy=[(0, 0), (32, 32)], fill=255)
        draw.rectangle(xy=[(32, 32), (64, 64)], fill=255)

        out_png = tmp_path / "temp.png"
        im.save(out_png, transparency=0)
        with Image.open(out_png) as im_png:
            out_avif = tmp_path / "temp.avif"
            im_png.save(out_avif, quality=100)

            with Image.open(out_avif) as expected:
                assert_image_similar(im_png.convert("RGBA"), expected, 0.17)

    def test_decoder_strict_flags(self) -> None:
        # This would fail if full avif strictFlags were enabled
        with Image.open("Tests/images/avif/hopper-missing-pixi.avif") as im:
            assert im.size == (128, 128)

    @skip_unless_avif_encoder("aom")
    @pytest.mark.parametrize("speed", [-1, 1, 11])
    def test_aom_optimizations(self, tmp_path: Path, speed: int) -> None:
        test_file = tmp_path / "temp.avif"
        hopper().save(test_file, codec="aom", speed=speed)

    @skip_unless_avif_encoder("svt")
    def test_svt_optimizations(self, tmp_path: Path) -> None:
        test_file = tmp_path / "temp.avif"
        hopper().save(test_file, codec="svt", speed=1)


@skip_unless_feature("avif")
class TestAvifAnimation:
    @contextmanager
    def star_frames(self) -> Generator[list[Image.Image], None, None]:
        with Image.open("Tests/images/avif/star.png") as f:
            yield [f, f.rotate(90), f.rotate(180), f.rotate(270)]

    def test_n_frames(self) -> None:
        """
        Ensure that AVIF format sets n_frames and is_animated attributes
        correctly.
        """

        with Image.open(TEST_AVIF_FILE) as im:
            assert isinstance(im, AvifImagePlugin.AvifImageFile)
            assert im.n_frames == 1
            assert not im.is_animated

        with Image.open("Tests/images/avif/star.avifs") as im:
            assert isinstance(im, AvifImagePlugin.AvifImageFile)
            assert im.n_frames == 5
            assert im.is_animated

    def test_write_animation_P(self, tmp_path: Path) -> None:
        """
        Convert an animated GIF to animated AVIF, then compare the frame
        count, and ensure the frames are visually similar to the originals.
        """

        with Image.open("Tests/images/avif/star.gif") as original:
            assert isinstance(original, GifImagePlugin.GifImageFile)
            assert original.n_frames > 1

            temp_file = tmp_path / "temp.avif"
            original.save(temp_file, save_all=True)
            with Image.open(temp_file) as im:
                assert isinstance(im, AvifImagePlugin.AvifImageFile)
                assert im.n_frames == original.n_frames

                # Compare first frame in P mode to frame from original GIF
                assert_image_similar(im, original.convert("RGBA"), 2)

                # Compare later frames in RGBA mode to frames from original GIF
                for frame in range(1, original.n_frames):
                    original.seek(frame)
                    im.seek(frame)
                    assert_image_similar(im, original, 2.54)

    def test_write_animation_RGBA(self, tmp_path: Path) -> None:
        """
        Write an animated AVIF from RGBA frames, and ensure the frames
        are visually similar to the originals.
        """

        def check(temp_file: Path) -> None:
            with Image.open(temp_file) as im:
                assert isinstance(im, AvifImagePlugin.AvifImageFile)
                assert im.n_frames == 4

                # Compare first frame to original
                assert_image_similar(im, frame1, 2.7)

                # Compare second frame to original
                im.seek(1)
                assert_image_similar(im, frame2, 4.1)

        with self.star_frames() as frames:
            frame1 = frames[0]
            frame2 = frames[1]
            temp_file1 = tmp_path / "temp.avif"
            frames[0].copy().save(temp_file1, save_all=True, append_images=frames[1:])
            check(temp_file1)

            # Test appending using a generator
            def imGenerator(
                ims: list[Image.Image],
            ) -> Generator[Image.Image, None, None]:
                yield from ims

            temp_file2 = tmp_path / "temp_generator.avif"
            frames[0].copy().save(
                temp_file2,
                save_all=True,
                append_images=imGenerator(frames[1:]),
            )
            check(temp_file2)

    def test_sequence_dimension_mismatch_check(self, tmp_path: Path) -> None:
        temp_file = tmp_path / "temp.avif"
        frame1 = Image.new("RGB", (100, 100))
        frame2 = Image.new("RGB", (150, 150))
        with pytest.raises(ValueError):
            frame1.save(temp_file, save_all=True, append_images=[frame2])

    def test_heif_raises_unidentified_image_error(self) -> None:
        with pytest.raises(UnidentifiedImageError):
            with Image.open("Tests/images/avif/hopper.heif"):
                pass

    @pytest.mark.parametrize("alpha_premultiplied", [False, True])
    def test_alpha_premultiplied(
        self, tmp_path: Path, alpha_premultiplied: bool
    ) -> None:
        temp_file = tmp_path / "temp.avif"
        color = (200, 200, 200, 1)
        im = Image.new("RGBA", (1, 1), color)
        im.save(temp_file, alpha_premultiplied=alpha_premultiplied)

        expected = (255, 255, 255, 1) if alpha_premultiplied else color
        with Image.open(temp_file) as reloaded:
            assert reloaded.getpixel((0, 0)) == expected

    def test_timestamp_and_duration(self, tmp_path: Path) -> None:
        """
        Try passing a list of durations, and make sure the encoded
        timestamps and durations are correct.
        """

        durations = [1, 10, 20, 30, 40]
        temp_file = tmp_path / "temp.avif"
        with self.star_frames() as frames:
            frames[0].save(
                temp_file,
                save_all=True,
                append_images=(frames[1:] + [frames[0]]),
                duration=durations,
            )

        with Image.open(temp_file) as im:
            assert isinstance(im, AvifImagePlugin.AvifImageFile)
            assert im.n_frames == 5
            assert im.is_animated

            # Check that timestamps and durations match original values specified
            timestamp = 0
            for frame in range(im.n_frames):
                im.seek(frame)
                im.load()
                assert im.info["duration"] == durations[frame]
                assert im.info["timestamp"] == timestamp
                timestamp += durations[frame]

    def test_seeking(self, tmp_path: Path) -> None:
        """
        Create an animated AVIF file, and then try seeking through frames in
        reverse-order, verifying the timestamps and durations are correct.
        """

        duration = 33
        temp_file = tmp_path / "temp.avif"
        with self.star_frames() as frames:
            frames[0].save(
                temp_file,
                save_all=True,
                append_images=(frames[1:] + [frames[0]]),
                duration=duration,
            )

        with Image.open(temp_file) as im:
            assert isinstance(im, AvifImagePlugin.AvifImageFile)
            assert im.n_frames == 5
            assert im.is_animated

            # Traverse frames in reverse, checking timestamps and durations
            timestamp = duration * (im.n_frames - 1)
            for frame in reversed(range(im.n_frames)):
                im.seek(frame)
                im.load()
                assert im.info["duration"] == duration
                assert im.info["timestamp"] == timestamp
                timestamp -= duration

    def test_seek_errors(self) -> None:
        with Image.open("Tests/images/avif/star.avifs") as im:
            with pytest.raises(EOFError):
                im.seek(-1)

            with pytest.raises(EOFError):
                im.seek(42)


MAX_THREADS = os.cpu_count() or 1


@skip_unless_feature("avif")
class TestAvifLeaks(PillowLeakTestCase):
    mem_limit = MAX_THREADS * 3 * 1024
    iterations = 100

    @pytest.mark.skipif(
        is_docker_qemu(), reason="Skipping on cross-architecture containers"
    )
    def test_leak_load(self) -> None:
        with open(TEST_AVIF_FILE, "rb") as f:
            im_data = f.read()

        def core() -> None:
            with Image.open(BytesIO(im_data)) as im:
                im.load()
            gc.collect()

        self._test_leak(core)



from __future__ import annotations

from pathlib import Path, PurePath

import pytest

from PIL import _util


@pytest.mark.parametrize(
    "test_path", ["filename.ext", Path("filename.ext"), PurePath("filename.ext")]
)
def test_is_path(test_path: str | Path | PurePath) -> None:
    # Act
    it_is = _util.is_path(test_path)

    # Assert
    assert it_is


def test_is_not_path(tmp_path: Path) -> None:
    # Arrange
    with (tmp_path / "temp.ext").open("w") as fp:
        pass

    # Act
    it_is_not = _util.is_path(fp)

    # Assert
    assert not it_is_not


def test_deferred_error() -> None:
    # Arrange

    # Act
    thing = _util.DeferredError.new(ValueError("Some error text"))

    # Assert
    with pytest.raises(ValueError):
        thing.some_attr



from __future__ import annotations

import logging
import os

import pytest

from PIL import Image

from .helper import hopper


def test_sanity() -> None:
    im = hopper()
    px = im.load()

    assert px is not None
    assert px[0, 0] == (20, 20, 70)


def test_close() -> None:
    im = Image.open("Tests/images/hopper.gif")
    im.close()
    with pytest.raises(ValueError):
        im.load()
    with pytest.raises(ValueError):
        im.getpixel((0, 0))


def test_close_after_load(caplog: pytest.LogCaptureFixture) -> None:
    im = Image.open("Tests/images/hopper.gif")
    im.load()
    with caplog.at_level(logging.DEBUG):
        im.close()
    assert len(caplog.records) == 0


def test_contextmanager() -> None:
    fn = None
    with Image.open("Tests/images/hopper.gif") as im:
        assert im.fp is not None
        fn = im.fp.fileno()
        os.fstat(fn)

    with pytest.raises(OSError):
        os.fstat(fn)


def test_contextmanager_non_exclusive_fp() -> None:
    with open("Tests/images/hopper.gif", "rb") as fp:
        with Image.open(fp):
            pass

        assert not fp.closed



from __future__ import annotations

import pytest

from PIL import Image, ImageDraw, ImageFont, ImageText, features

from .helper import assert_image_similar_tofile, skip_unless_feature

FONT_PATH = "Tests/fonts/FreeMono.ttf"


@pytest.fixture(
    scope="module",
    params=[
        pytest.param(ImageFont.Layout.BASIC),
        pytest.param(ImageFont.Layout.RAQM, marks=skip_unless_feature("raqm")),
    ],
)
def layout_engine(request: pytest.FixtureRequest) -> ImageFont.Layout:
    return request.param


@pytest.fixture(
    scope="module",
    params=[
        None,
        pytest.param(ImageFont.Layout.BASIC, marks=skip_unless_feature("freetype2")),
        pytest.param(ImageFont.Layout.RAQM, marks=skip_unless_feature("raqm")),
    ],
)
def font(
    request: pytest.FixtureRequest,
) -> ImageFont.ImageFont | ImageFont.FreeTypeFont:
    layout_engine = request.param
    if layout_engine is None:
        return ImageFont.load_default_imagefont()
    else:
        return ImageFont.truetype(FONT_PATH, 20, layout_engine=layout_engine)


def test_get_length(font: ImageFont.ImageFont | ImageFont.FreeTypeFont) -> None:
    factor = 1 if isinstance(font, ImageFont.ImageFont) else 2
    assert ImageText.Text("A", font).get_length() == 6 * factor
    assert ImageText.Text("AB", font).get_length() == 12 * factor
    assert ImageText.Text("M", font).get_length() == 6 * factor
    assert ImageText.Text("y", font).get_length() == 6 * factor
    assert ImageText.Text("a", font).get_length() == 6 * factor

    text = ImageText.Text("\n", font)
    with pytest.raises(ValueError, match="can't measure length of multiline text"):
        text.get_length()


@pytest.mark.parametrize(
    "text, expected",
    (
        ("A", (0, 4, 12, 16)),
        ("AB", (0, 4, 24, 16)),
        ("M", (0, 4, 12, 16)),
        ("y", (0, 7, 12, 20)),
        ("a", (0, 7, 12, 16)),
    ),
)
def test_get_bbox(
    font: ImageFont.ImageFont | ImageFont.FreeTypeFont,
    text: str,
    expected: tuple[int, int, int, int],
) -> None:
    if isinstance(font, ImageFont.ImageFont):
        expected = (0, 0, expected[2] // 2, 11)
    assert ImageText.Text(text, font).get_bbox() == expected


def test_standard_embedded_color(layout_engine: ImageFont.Layout) -> None:
    if features.check_module("freetype2"):
        font = ImageFont.truetype(FONT_PATH, 40, layout_engine=layout_engine)
        text = ImageText.Text("Hello World!", font)
        text.embed_color()
        assert text.get_length() == 288

        im = Image.new("RGB", (300, 64), "white")
        draw = ImageDraw.Draw(im)
        draw.text((10, 10), text, "#fa6")

        assert_image_similar_tofile(im, "Tests/images/standard_embedded.png", 3.1)

    text = ImageText.Text("", mode="1")
    with pytest.raises(
        ValueError, match="Embedded color supported only in RGB and RGBA modes"
    ):
        text.embed_color()


@skip_unless_feature("freetype2")
def test_stroke() -> None:
    for suffix, stroke_fill in {"same": None, "different": "#0f0"}.items():
        # Arrange
        im = Image.new("RGB", (120, 130))
        draw = ImageDraw.Draw(im)
        font = ImageFont.truetype(FONT_PATH, 120)
        text = ImageText.Text("A", font)
        text.stroke(2, stroke_fill)

        # Act
        draw.text((12, 12), text, "#f00")

        # Assert
        assert_image_similar_tofile(
            im, "Tests/images/imagedraw_stroke_" + suffix + ".png", 3.1
        )


@pytest.mark.parametrize(
    "data, width, expected",
    (
        ("Hello World!", 100, "Hello World!"),  # No wrap required
        ("Hello World!", 50, "Hello\nWorld!"),  # Wrap word to a new line
        # Keep multiple spaces within a line
        ("Keep  multiple spaces", 90, "Keep  multiple\nspaces"),
        (" Keep\n leading space", 100, " Keep\n leading space"),
    ),
)
@pytest.mark.parametrize("string", (True, False))
def test_wrap(data: str, width: int, expected: str, string: bool) -> None:
    if string:
        text = ImageText.Text(data)
        assert text.wrap(width) is None
        assert text.text == expected
    else:
        text_bytes = ImageText.Text(data.encode())
        assert text_bytes.wrap(width) is None
        assert text_bytes.text == expected.encode()


def test_wrap_long_word() -> None:
    text = ImageText.Text("Hello World!")
    with pytest.raises(ValueError, match="Word does not fit within line"):
        text.wrap(25)


def test_wrap_unsupported(font: ImageFont.FreeTypeFont) -> None:
    transposed_font = ImageFont.TransposedFont(font)
    text = ImageText.Text("Hello World!", transposed_font)
    with pytest.raises(ValueError, match="TransposedFont not supported"):
        text.wrap(50)

    text = ImageText.Text("Hello World!", direction="ttb")
    with pytest.raises(ValueError, match="Only ltr direction supported"):
        text.wrap(50)


def test_wrap_height() -> None:
    width = 50 if features.check_module("freetype2") else 60
    text = ImageText.Text("Text does not fit within height")
    wrapped = text.wrap(width, 25 if features.check_module("freetype2") else 40)
    assert wrapped is not None
    assert wrapped.text == " within height"
    assert text.text == "Text does\nnot fit"

    text = ImageText.Text("Text does not fit\nwithin height")
    wrapped = text.wrap(width, 20)
    assert wrapped is not None
    assert wrapped.text == " not fit\nwithin height"
    assert text.text == "Text does"

    text = ImageText.Text("Text does not fit\n\nwithin height")
    wrapped = text.wrap(width, 25 if features.check_module("freetype2") else 40)
    assert wrapped is not None
    assert wrapped.text == "\nwithin height"
    assert text.text == "Text does\nnot fit"


def test_wrap_scaling_unsupported() -> None:
    font = ImageFont.load_default_imagefont()
    text = ImageText.Text("Hello World!", font)
    with pytest.raises(ValueError, match="'scaling' only supports FreeTypeFont"):
        text.wrap(50, scaling="shrink")

    if features.check_module("freetype2"):
        text = ImageText.Text("Hello World!")
        with pytest.raises(ValueError, match="'scaling' requires 'height'"):
            text.wrap(50, scaling="shrink")


@skip_unless_feature("freetype2")
def test_wrap_shrink() -> None:
    # No scaling required
    text = ImageText.Text("Hello World!")
    assert isinstance(text.font, ImageFont.FreeTypeFont)
    assert text.font.size == 10
    assert text.wrap(50, 50, "shrink") is None
    assert isinstance(text.font, ImageFont.FreeTypeFont)
    assert text.font.size == 10

    with pytest.raises(ValueError, match="Text could not be scaled"):
        text.wrap(50, 15, ("shrink", 9))

    assert text.wrap(50, 15, "shrink") is None
    assert text.font.size == 8

    text = ImageText.Text("Hello World!")
    assert text.wrap(50, 15, ("shrink", 7)) is None
    assert isinstance(text.font, ImageFont.FreeTypeFont)
    assert text.font.size == 8


@skip_unless_feature("freetype2")
def test_wrap_grow() -> None:
    # No scaling required
    text = ImageText.Text("Hello World!")
    assert isinstance(text.font, ImageFont.FreeTypeFont)
    assert text.font.size == 10
    assert text.wrap(58, 10, "grow") is None
    assert isinstance(text.font, ImageFont.FreeTypeFont)
    assert text.font.size == 10

    with pytest.raises(ValueError, match="Text could not be scaled"):
        text.wrap(50, 50, ("grow", 12))

    assert text.wrap(50, 50, "grow") is None
    assert text.font.size == 16

    text = ImageText.Text("A\nB")
    with pytest.raises(ValueError, match="Text could not be scaled"):
        text.wrap(50, 10, "grow")

    text = ImageText.Text("Hello World!")
    assert text.wrap(50, 50, ("grow", 18)) is None
    assert isinstance(text.font, ImageFont.FreeTypeFont)
    assert text.font.size == 16



from __future__ import annotations

import io
import re

import pytest

from PIL import features

from .helper import skip_unless_feature

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable


def test_check() -> None:
    # Check the correctness of the convenience function
    for module in features.modules:
        assert features.check_module(module) == features.check(module)
    for codec in features.codecs:
        assert features.check_codec(codec) == features.check(codec)
    for feature in features.features:
        assert features.check_feature(feature) == features.check(feature)


def test_version() -> None:
    # Check the correctness of the convenience function
    # and the format of version numbers

    def test(name: str, function: Callable[[str], str | None]) -> None:
        version = features.version(name)
        if not features.check(name):
            assert version is None
        else:
            assert function(name) == version
            if name != "PIL":
                if version is not None:
                    if name == "zlib" and features.check_feature("zlib_ng"):
                        version = re.sub(".zlib-ng$", "", version)
                    elif name == "libtiff":
                        version = re.sub("t$", "", version)
                assert version is None or re.search(r"\d+(\.\d+)*$", version)

    for module in features.modules:
        test(module, features.version_module)
    for codec in features.codecs:
        test(codec, features.version_codec)
    for feature in features.features:
        test(feature, features.version_feature)


@skip_unless_feature("libjpeg_turbo")
def test_libjpeg_turbo_version() -> None:
    version = features.version("libjpeg_turbo")
    assert version is not None
    assert re.search(r"\d+\.\d+\.\d+$", version)


@skip_unless_feature("libimagequant")
def test_libimagequant_version() -> None:
    version = features.version("libimagequant")
    assert version is not None
    assert re.search(r"\d+\.\d+\.\d+$", version)


@pytest.mark.parametrize("feature", features.modules)
def test_check_modules(feature: str) -> None:
    assert features.check_module(feature) in [True, False]


@pytest.mark.parametrize("feature", features.codecs)
def test_check_codecs(feature: str) -> None:
    assert features.check_codec(feature) in [True, False]


def test_check_warns_on_nonexistent() -> None:
    with pytest.warns(UserWarning, match="Unknown feature 'typo'."):
        has_feature = features.check("typo")
    assert has_feature is False


def test_supported_modules() -> None:
    assert isinstance(features.get_supported_modules(), list)
    assert isinstance(features.get_supported_codecs(), list)
    assert isinstance(features.get_supported_features(), list)
    assert isinstance(features.get_supported(), list)


def test_unsupported_codec() -> None:
    # Arrange
    codec = "unsupported_codec"
    # Act / Assert
    with pytest.raises(ValueError):
        features.check_codec(codec)
    with pytest.raises(ValueError):
        features.version_codec(codec)


def test_unsupported_module() -> None:
    # Arrange
    module = "unsupported_module"
    # Act / Assert
    with pytest.raises(ValueError):
        features.check_module(module)
    with pytest.raises(ValueError):
        features.version_module(module)


def test_unsupported_feature() -> None:
    # Arrange
    feature = "unsupported_feature"
    # Act / Assert
    with pytest.raises(ValueError):
        features.check_feature(feature)
    with pytest.raises(ValueError):
        features.version_feature(feature)


def test_unsupported_version() -> None:
    assert features.version("unsupported_version") is None


def test_modulenotfound(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(features, "features", {"test": ("PIL._test", "", "")})
    assert features.check_feature("test") is None


@pytest.mark.parametrize("supported_formats", (True, False))
def test_pilinfo(supported_formats: bool) -> None:
    buf = io.StringIO()
    features.pilinfo(buf, supported_formats=supported_formats)
    out = buf.getvalue()
    lines = out.splitlines()
    assert lines[0] == "-" * 68
    assert lines[1].startswith("Pillow ")
    assert lines[2].startswith("Python ")
    lines = lines[3:]
    while lines[0].startswith("    "):
        lines = lines[1:]
    assert lines[0] == "-" * 68
    assert lines[1].startswith("Python executable is")
    lines = lines[2:]
    if lines[0].startswith("Environment Python files loaded from"):
        lines = lines[1:]
    assert lines[0].startswith("System Python files loaded from")
    assert lines[1] == "-" * 68
    assert lines[2].startswith("Python Pillow modules loaded from ")
    assert lines[3].startswith("Binary Pillow modules loaded from ")
    assert lines[4] == "-" * 68
    jpeg = (
        "\n"
        + "-" * 68
        + "\n"
        + "JPEG image/jpeg\n"
        + "Extensions: .jfif, .jpe, .jpeg, .jpg\n"
        + "Features: open, save\n"
        + "-" * 68
        + "\n"
    )
    assert supported_formats == (jpeg in out)



#
# Python Imaging Library
# $Id$
#
# stuff to read simple, teragon-style palette files
#
# History:
#       97-08-23 fl     Created
#
# Copyright (c) Secret Labs AB 1997.
# Copyright (c) Fredrik Lundh 1997.
#
# See the README file for information on usage and redistribution.
#
from __future__ import annotations

from typing import IO

from ._binary import o8


class PaletteFile:
    """File handler for Teragon-style palette files."""

    rawmode = "RGB"

    def __init__(self, fp: IO[bytes]) -> None:
        palette = [o8(i) * 3 for i in range(256)]

        while True:
            s = fp.readline()

            if not s:
                break
            if s.startswith(b"#"):
                continue
            if len(s) > 100:
                msg = "bad palette file"
                raise SyntaxError(msg)

            v = [int(x) for x in s.split()]
            try:
                [i, r, g, b] = v
            except ValueError:
                [i, r] = v
                g = b = r

            if 0 <= i <= 255:
                palette[i] = o8(r) + o8(g) + o8(b)

        self.palette = b"".join(palette)

    def getpalette(self) -> tuple[bytes, str]:
        return self.palette, self.rawmode



from __future__ import annotations

from io import BytesIO

import pytest

from PIL import CurImagePlugin, Image
from PIL._binary import o8
from PIL._binary import o16le as o16
from PIL._binary import o32le as o32

TEST_FILE = "Tests/images/deerstalker.cur"


def test_sanity() -> None:
    with Image.open(TEST_FILE) as im:
        assert im.size == (32, 32)
        assert isinstance(im, CurImagePlugin.CurImageFile)
        # Check some pixel colors to ensure image is loaded properly
        assert im.getpixel((10, 1)) == (0, 0, 0, 0)
        assert im.getpixel((11, 1)) == (253, 254, 254, 1)
        assert im.getpixel((16, 16)) == (84, 87, 86, 255)


def test_largest_cursor() -> None:
    magic = b"\x00\x00\x02\x00"
    sizes = ((1, 1), (8, 8), (4, 4))
    data = magic + o16(len(sizes))
    for w, h in sizes:
        image_offset = 6 + len(sizes) * 16 if (w, h) == max(sizes) else 0
        data += o8(w) + o8(h) + o8(0) * 10 + o32(image_offset)
    data += (
        o32(12)  # header size
        + o16(8)  # width
        + o16(16)  # height
        + o16(0)  # planes
        + o16(1)  # bits
    )
    with Image.open(BytesIO(data)) as im:
        assert im.size == (8, 8)


@pytest.mark.parametrize("mode", ("1", "L"))
def test_mask(mode: str) -> None:
    with Image.open("Tests/images/mask_" + mode + ".cur") as im:
        assert im.mode == "LA"

        for i, value in enumerate(
            [
                (0, 255),  # AND 0 XOR 0 is black
                (255, 255),  # AND 0 XOR 1 is white
                (0, 0),  # AND 1 XOR 0 is transparent
                (0, 0),  # AND 1 XOR 1 is transparent
            ]
        ):
            assert im.getpixel((0, i)) == value


def test_invalid_file() -> None:
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        CurImagePlugin.CurImageFile(invalid_file)

    no_cursors_file = "Tests/images/no_cursors.cur"

    cur = CurImagePlugin.CurImageFile(TEST_FILE)
    assert cur.fp is not None
    cur.fp.close()
    with open(no_cursors_file, "rb") as cur.fp:
        with pytest.raises(TypeError):
            cur._open()



from __future__ import annotations

from io import BytesIO

from PIL import Image

from .helper import PillowLeakTestCase, skip_unless_feature

test_file = "Tests/images/hopper.webp"


@skip_unless_feature("webp")
class TestWebPLeaks(PillowLeakTestCase):
    mem_limit = 3 * 1024  # kb
    iterations = 100

    def test_leak_load(self) -> None:
        with open(test_file, "rb") as f:
            im_data = f.read()

        def core() -> None:
            with Image.open(BytesIO(im_data)) as im:
                im.load()

        self._test_leak(core)



"""
pytest-benchmark tests for Pillow features.
"""

from __future__ import annotations

import pathlib
from importlib.util import find_spec
from io import BytesIO

import pytest

from PIL import Image, ImageChops, ImageFilter
from PIL.Image import Resampling, Transpose

TYPE_CHECKING = False
if TYPE_CHECKING:
    from collections.abc import Callable

    from pytest_benchmark.fixture import (  # type: ignore[import-not-found]
        BenchmarkFixture,
    )

if not (find_spec("pytest_benchmark") or find_spec("pytest_codspeed")):
    pytest.skip("pytest-benchmark or pytest-codspeed required", allow_module_level=True)

# These can be adjusted to add more modes to benchmark
# (however all features benchmarked might not support all PIL modes).
MODES = ["RGB", "RGBA", "L", "LA"]

# The size for generated test images.
# Note that adjusting this will naturally change how long operations take.
# The `bench` fixture takes care of saving this information in the extra info
# for the benchmark run, so that throughput (Mpx/s) can be recomputed in the future.
SIZES = [(1024, 1024)]

# For benchmarks that act on test fixture files, these are the paths loaded.
IMAGES_PATH = pathlib.Path(__file__).parent / "images"
PATHS = [
    IMAGES_PATH / "flower2.jpg",
]

# These are derived from the other configuration, above.
RGB_MODES = [mode for mode in MODES if mode.startswith("RGB")]
ALPHA_MODES = [mode for mode in MODES if mode.endswith("A")]
SCALE_MODES = [*MODES, "I", "F"]


def _format_size(size: tuple[int, int]) -> str:
    return f"{size[0]}x{size[1]}"


def _format_path(path: pathlib.Path) -> str:
    return path.name


@pytest.fixture
def bench(
    request: pytest.FixtureRequest,
    benchmark: BenchmarkFixture,
) -> BenchmarkFixture:
    """
    pytest-benchmark with extra information.
    """
    try:
        benchmark.extra_info["mode"] = request.getfixturevalue("mode")
    except LookupError:
        pass
    try:
        size = request.getfixturevalue("size")
        benchmark.extra_info["size"] = _format_size(size)
        benchmark.extra_info["pixels"] = size[0] * size[1]
    except LookupError:
        pass
    return benchmark


def make_pillow_image(
    mode: str,
    size: tuple[int, int],
    pattern_offset: int = 0,
) -> Image.Image:
    im = Image.new("RGB", size)
    n = im.width * im.height * 3
    period = bytes((i + pattern_offset) % 256 for i in range(256))
    im.frombytes((period * (n // 256 + 1))[:n])
    return im.convert(mode)


@pytest.mark.benchmark(group="composition")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_blend(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
) -> None:
    im1 = make_pillow_image(mode, size)
    im2 = make_pillow_image(mode, size, pattern_offset=1024)
    result = bench(Image.blend, im1, im2, 0.5)
    assert result.size == im1.size


@pytest.mark.benchmark(group="scale")
@pytest.mark.parametrize("resampler", Resampling, ids=lambda r: r.name)
@pytest.mark.parametrize("scale", [0.01, 0.125, 0.8, 2.14])
@pytest.mark.parametrize("mode", SCALE_MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_scale(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    scale: float,
    resampler: Resampling,
) -> None:
    im = make_pillow_image(mode, size)
    dest = (round(scale * im.width), round(scale * im.height))
    bench.extra_info["label"] = [f"{dest[0]}x{dest[1]}", resampler.name]
    bench(im.resize, dest, resampler)


@pytest.mark.benchmark(group="blur")
@pytest.mark.parametrize("radius", [1, 10, 30])
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_box_blur(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    radius: int,
) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [f"{radius}px"]
    bench(im.filter, ImageFilter.BoxBlur(radius))


@pytest.mark.benchmark(group="composition")
@pytest.mark.parametrize("alpha", ["opaque", "transparent", "mixed"])
@pytest.mark.parametrize("mode", ALPHA_MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_alpha_composite(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    alpha: str,
) -> None:
    im1 = make_pillow_image(mode, size)
    im2 = make_pillow_image(mode, size, pattern_offset=1024)
    if alpha == "opaque":
        im2.putalpha(255)
    elif alpha == "transparent":
        im2.putalpha(0)
    else:  # "mixed"
        width, height = size
        gradient = bytes((x * 256) // width for x in range(width)) * height
        im2.putalpha(Image.frombytes("L", size, gradient))
    bench.extra_info["label"] = ["Composition", alpha]
    result = bench(Image.alpha_composite, im1, im2)
    assert result.size == im1.size


@pytest.mark.benchmark(group="convert")
@pytest.mark.parametrize(
    "mode_from, mode_to",
    [
        ("RGB", "L"),
        ("RGBA", "LA"),
        ("RGBa", "RGBA"),
        ("RGBA", "RGBa"),
    ],
)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_convert(
    bench: BenchmarkFixture,
    mode_from: str,
    mode_to: str,
    size: tuple[int, int],
) -> None:
    im = make_pillow_image(mode_from, size)
    bench.extra_info["label"] = [f"{mode_from} to {mode_to}"]
    bench(im.convert, mode_to)


@pytest.mark.benchmark(group="crop")
@pytest.mark.parametrize(
    "scale",
    [
        (0.9, 0.9),
        (1.1, 1.1),
        (1.1, 0.9),
    ],
)
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_crop(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    scale: tuple[float, float],
) -> None:
    im = make_pillow_image(mode, size)
    w, h = im.size
    width, height = round(scale[0] * w), round(scale[1] * h)
    left = (w - width) // 2
    top = (h - height) // 2
    box = (left, top, left + width, top + height)
    bench.extra_info["label"] = [f"{width}x{height}"]
    bench(im.crop, box)


@pytest.mark.benchmark(group="filter")
@pytest.mark.parametrize(
    "filter",
    [
        ImageFilter.SMOOTH,
        ImageFilter.SHARPEN,
        ImageFilter.SMOOTH_MORE,
    ],
    ids=lambda f: f.name,
)
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_filter(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    filter: type[ImageFilter.BuiltinFilter],
) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [filter.name]
    bench(im.filter, filter)


@pytest.mark.benchmark(group="lut")
@pytest.mark.parametrize(
    "channels, table_size",
    [
        (3, 4),
        (3, 16),
        (3, 36),
        (4, 4),
        (4, 16),
        (4, 36),
    ],
)
@pytest.mark.parametrize("mode", RGB_MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_lut(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    channels: int,
    table_size: int,
) -> None:
    im = make_pillow_image(mode, size)
    if channels == 3:
        lut = ImageFilter.Color3DLUT.generate(
            table_size, lambda r, g, b: (r, g, b), channels, "RGB"
        )
    else:
        lut = ImageFilter.Color3DLUT.generate(
            table_size, lambda r, g, b: (r, g, b, r), channels, "RGBA"
        )

    bench.extra_info["label"] = [f"{table_size}³ table to {channels}D"]
    bench(im.filter, lut)


@pytest.mark.benchmark(group="rotate_right")
@pytest.mark.parametrize("op", Transpose, ids=lambda t: t.name)
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_rotate_right(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    op: Transpose,
) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [op.name]
    bench(im.transpose, op)


@pytest.mark.benchmark(group="load")
@pytest.mark.parametrize("path", PATHS, ids=_format_path)
def test_load(bench: BenchmarkFixture, path: pathlib.Path) -> None:
    def run() -> None:
        with Image.open(path) as im:
            im.load()

    bench(run)


@pytest.mark.benchmark(group="save")
@pytest.mark.parametrize("path", PATHS, ids=_format_path)
def test_save_jpeg(bench: BenchmarkFixture, path: pathlib.Path) -> None:
    with Image.open(path) as im:
        im.load()
    bench(lambda: im.save(BytesIO(), format="JPEG", quality=85))


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_allocate(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    bench.extra_info["label"] = [f"mode {mode}"]
    bench(Image.new, mode, size)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_unpack(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    data = im.tobytes()
    bench.extra_info["label"] = [f"Unpack from {mode}"]
    bench(im.frombytes, data)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_pack(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [f"Pack to {mode}"]
    bench(im.tobytes)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_split(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [f"split {mode}"]
    bench(im.split)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_getband(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    band = len(im.getbands()) - 1
    bench.extra_info["label"] = [f"get {mode[band]} of {mode}"]
    bench(im.getchannel, band)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_merge(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bands = im.split()
    bench.extra_info["label"] = [f"merge {mode}"]
    bench(Image.merge, mode, bands)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_fill(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    nbands = len(Image.new(mode, (1, 1)).getbands())
    color = (10, 20, 30, 40)[:nbands] if nbands > 1 else 10
    bench.extra_info["label"] = [f"fill {mode}"]
    bench(Image.new, mode, size, color)


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", ["1", "F", "I", "L", "P"])
def test_linear_gradient(bench: BenchmarkFixture, mode: str) -> None:
    result = bench(Image.linear_gradient, mode)
    assert result.size == (256, 256)
    assert result.mode == mode


@pytest.mark.benchmark(group="allocate")
@pytest.mark.parametrize("mode", ["1", "F", "I", "L", "P"])
def test_radial_gradient(bench: BenchmarkFixture, mode: str) -> None:
    result = bench(Image.radial_gradient, mode)
    assert result.size == (256, 256)
    assert result.mode == mode


CHOPS_OPS = [
    ImageChops.add,
    ImageChops.subtract,
    ImageChops.multiply,
    ImageChops.screen,
    ImageChops.difference,
    ImageChops.lighter,
    ImageChops.darker,
    ImageChops.add_modulo,
    ImageChops.soft_light,
    ImageChops.hard_light,
    ImageChops.overlay,
]


@pytest.mark.benchmark(group="chops")
@pytest.mark.parametrize("op", CHOPS_OPS, ids=lambda f: f.__name__)
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_chops(
    bench: BenchmarkFixture,
    mode: str,
    size: tuple[int, int],
    op: Callable[[Image.Image, Image.Image], Image.Image],
) -> None:
    im1 = make_pillow_image(mode, size)
    im2 = make_pillow_image(mode, size, pattern_offset=1024)
    bench.extra_info["label"] = [op.__name__]
    result = bench(op, im1, im2)
    assert result.size == im1.size


@pytest.mark.benchmark(group="chops")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_invert(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = ["invert"]
    bench(ImageChops.invert, im)


@pytest.mark.benchmark(group="chops")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_offset(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = ["offset"]
    bench(ImageChops.offset, im, 123, 45)


@pytest.mark.benchmark(group="histogram")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_histogram(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [f"histogram {mode}"]
    bench(im.histogram)


@pytest.mark.benchmark(group="histogram")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_histogram_masked(
    bench: BenchmarkFixture, mode: str, size: tuple[int, int]
) -> None:
    im = make_pillow_image(mode, size)
    mask = make_pillow_image("L", size)
    bench.extra_info["label"] = [f"masked histogram {mode}"]
    bench(im.histogram, mask)


L_MATRIX = (0.299, 0.587, 0.114, 0.0)
RGB_MATRIX = (
    0.412, 0.357, 0.180, 0.0,
    0.212, 0.715, 0.072, 0.0,
    0.019, 0.119, 0.950, 0.0,
)  # fmt: skip


@pytest.mark.benchmark(group="convert")
@pytest.mark.parametrize(
    "mode_to, matrix",
    [("L", L_MATRIX), ("RGB", RGB_MATRIX)],
)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_matrix_convert(
    bench: BenchmarkFixture,
    mode_to: str,
    matrix: tuple[float, ...],
    size: tuple[int, int],
) -> None:
    im = make_pillow_image("RGB", size)
    bench.extra_info["label"] = [f"matrix RGB to {mode_to}"]
    bench(im.convert, mode_to, matrix)


@pytest.mark.benchmark(group="point")
@pytest.mark.parametrize("mode", MODES)
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_point_lut(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    lut = [255 - i for i in range(256)] * len(im.getbands())
    bench.extra_info["label"] = [f"LUT {mode}"]
    bench(im.point, lut)


@pytest.mark.benchmark(group="point")
@pytest.mark.parametrize("mode", ["I", "F"])
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_point_transform(
    bench: BenchmarkFixture, mode: str, size: tuple[int, int]
) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [f"transform {mode}"]
    bench(im.point, lambda v: v * 1.5 + 3.0)


@pytest.mark.benchmark(group="quantize")
@pytest.mark.parametrize("mode", [m for m in MODES if m in ("L", "RGB", "RGBA")])
@pytest.mark.parametrize("size", SIZES, ids=_format_size)
def test_quantize(bench: BenchmarkFixture, mode: str, size: tuple[int, int]) -> None:
    im = make_pillow_image(mode, size)
    bench.extra_info["label"] = [f"quantize {mode}"]
    result = bench(im.quantize, 256)
    assert result.mode == "P"



from __future__ import annotations

import pytest

from PIL import Image

from .helper import assert_image_equal, hopper


@pytest.mark.parametrize("mode", ("1", "P", "L", "RGB", "I", "F"))
def test_crop(mode: str) -> None:
    im = hopper(mode)
    assert_image_equal(im.crop(), im)

    cropped = im.crop((50, 50, 100, 100))
    assert cropped.mode == mode
    assert cropped.size == (50, 50)


def test_wide_crop() -> None:
    def crop(bbox: tuple[int, int, int, int]) -> tuple[int, ...]:
        i = im.crop(bbox)
        h = i.histogram()
        while h and not h[-1]:
            del h[-1]
        return tuple(h)

    im = Image.new("L", (100, 100), 1)

    assert crop((0, 0, 100, 100)) == (0, 10000)
    assert crop((25, 25, 75, 75)) == (0, 2500)

    # sides
    assert crop((-25, 0, 25, 50)) == (1250, 1250)
    assert crop((0, -25, 50, 25)) == (1250, 1250)
    assert crop((75, 0, 125, 50)) == (1250, 1250)
    assert crop((0, 75, 50, 125)) == (1250, 1250)

    assert crop((-25, 25, 125, 75)) == (2500, 5000)
    assert crop((25, -25, 75, 125)) == (2500, 5000)

    # corners
    assert crop((-25, -25, 25, 25)) == (1875, 625)
    assert crop((75, -25, 125, 25)) == (1875, 625)
    assert crop((75, 75, 125, 125)) == (1875, 625)
    assert crop((-25, 75, 25, 125)) == (1875, 625)


@pytest.mark.parametrize("box", ((8, 2, 2, 8), (2, 8, 8, 2), (8, 8, 2, 2)))
def test_negative_crop(box: tuple[int, int, int, int]) -> None:
    im = Image.new("RGB", (10, 10))

    with pytest.raises(ValueError):
        im.crop(box)


def test_crop_float() -> None:
    # Check cropping floats are rounded to nearest integer
    # https://github.com/python-pillow/Pillow/issues/1744

    # Arrange
    im = Image.new("RGB", (10, 10))
    assert im.size == (10, 10)

    # Act
    cropped = im.crop((0.9, 1.1, 4.2, 5.8))

    # Assert
    assert cropped.size == (3, 5)


def test_crop_crash() -> None:
    # Image.crop crashes prepatch with an access violation
    # apparently a use after free on Windows, see
    # https://github.com/python-pillow/Pillow/issues/1077

    test_img = "Tests/images/bmp/g/pal8-0.bmp"
    extents = (1, 1, 10, 10)
    # works prepatch
    with Image.open(test_img) as img:
        img1 = img.crop(extents)
    img1.load()

    # fail prepatch
    with Image.open(test_img) as img:
        img2 = img.crop(extents)
    img2.load()


def test_crop_zero() -> None:
    im = Image.new("RGB", (0, 0), "white")

    cropped = im.crop((0, 0, 0, 0))
    assert cropped.size == (0, 0)

    cropped = im.crop((10, 10, 20, 20))
    assert cropped.size == (10, 10)
    assert cropped.getpixel((0, 0)) == (0, 0, 0)

    im = Image.new("RGB", (0, 0))

    cropped = im.crop((10, 10, 20, 20))
    assert cropped.size == (10, 10)
    assert cropped.getpixel((2, 0)) == (0, 0, 0)
